/* termio.c */

/* Test termio ioctl() and signals */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <errno.h>
#include <signal.h>
#include <process.h>
#include <setjmp.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/termio.h>

#define FALSE 0
#define TRUE  1

static int jump_flag = FALSE;
static int nread_flag = FALSE;

static jmp_buf jump;

static void handler (int sig)
    {
    if (sig == SIGINT)
        {
        (void)printf ("(SIGINT)");
        (void)signal (sig, SIG_ACK);
        }
    else if (sig == SIGALRM)
        {
        (void)printf ("(SIGALRM)");
        (void)signal (sig, SIG_ACK);
        }
    else
        {
        (void)printf ("Signal %d received. Process stopped.\n", sig);
        exit (1);
        }
    if (jump_flag)
        longjmp (jump, 1);
    }

int main (int argc, char *argv[])
    {
    char buf[100], *p;
    int i, n, req;
    struct termio tio;

    for (i = 1; i < argc; ++i)
        if (strcmp (argv[i], "-j") == 0)
            jump_flag = TRUE;
        else if (strcmp (argv[i], "-n") == 0)
            nread_flag = TRUE;
        else
            {
            (void)puts ("Usage: termio [-j]");
            return (1);
            }
    if (setjmp (jump) != 0)
        (void)puts ("jumped");
    (void)signal (SIGINT, handler);
    (void)signal (SIGBREAK, handler);
    (void)signal (SIGQUIT, handler);
    (void)signal (SIGALRM, handler);
    (void)ioctl (0, TCGETA, &tio);
    tio.c_lflag &= ~IDEFAULT;
    (void)ioctl (0, TCSETA, &tio);
    req = sizeof (buf) - 1;
    for (;;)
        {
        if (nread_flag)
            {
            if (ioctl (0, FIONREAD, &n) == -1)
                perror ("ioctl FIONREAD");
            else
                (void)printf ("%d> ", n);
            }
        n = read (0, buf, req);
        if (n < 0)
            {
            if (errno == EAGAIN)
                {
                (void)puts ("No data available -- sleeping for 2 seconds");
                (void)sleep (2);
                }
            else
                perror ("read");
            }
        if (n >= 0)
            {
            (void)write (1, "<", 1);
            (void)write (1, buf, n);
            (void)write (1, ">\n", 2);
            }
        if (n > 0)
            {
            buf[n] = 0;
            p = strchr (buf, '\n');
            if (p != NULL) *p = 0;
            if (buf[0] == '?')
                {
                (void)puts ("?          Help");
                (void)puts ("$          Quit");
                (void)puts ("!CMD       Shell escape");
                (void)puts ("+          Toggle ICANON");
                (void)puts (":          Toggle IDEFAULT");
                (void)puts ("|          Toggle IDELETE");
                (void)puts (";          Toggle ECHO");
                (void)puts ("-          Toggle O_NDELAY");
                (void)puts ("%SEC       Set alarm clock");
                (void)puts ("#SEC       Sleep (append 'f' to flush buffer)");
                (void)puts ("=CHARS     Set number of characters to read");
                (void)puts (">CHARS     Set VMIN");
                (void)puts ("<TENTHS    Set VTIME");
                (void)puts (".SIG       Ignore signal");
                (void)puts ("*SIG       Handle signal");
                (void)puts ("^SIG       Set default processing for signal");
                (void)puts ("~SIG       Raise signal");
                }
            else if (buf[0] == '$')
                break;
            else if (buf[0] == '!')
                system (buf+1);
            else if (buf[0] == '%')
                (void)printf ("alarm=%u\n", alarm (atoi (buf+1)));
            else if (buf[0] == '#')
                {
                n = atoi (buf+1);
                if (n > 0)
                    (void)sleep (n);
                if (strchr (buf+1, 'f'))
                    (void)ioctl (0, TCFLSH, 0);
                }
            else if (buf[0] == '+')
                {
                tio.c_lflag ^= ICANON;
                (void)ioctl (0, TCSETA, &tio);
                }
            else if (buf[0] == ':')
                {
                tio.c_lflag ^= IDEFAULT;
                (void)ioctl (0, TCSETA, &tio);
                }
            else if (buf[0] == ';')
                {
                tio.c_lflag ^= ECHO;
                (void)ioctl (0, TCSETA, &tio);
                }
            else if (buf[0] == '|')
                {
                tio.c_iflag ^= IDELETE;
                (void)ioctl (0, TCSETA, &tio);
                }
            else if (buf[0] == '>')
                {
                tio.c_cc[VMIN] = (unsigned char)atoi (buf+1);
                (void)ioctl (0, TCSETA, &tio);
                }
            else if (buf[0] == '<')
                {
                tio.c_cc[VTIME] = (unsigned char)atoi (buf+1);
                (void)ioctl (0, TCSETA, &tio);
                }
            else if (buf[0] == '.')
                (void)signal (atoi (buf+1), SIG_IGN);
            else if (buf[0] == '*')
                (void)signal (atoi (buf+1), handler);
            else if (buf[0] == '^')
                (void)signal (atoi (buf+1), SIG_DFL);
            else if (buf[0] == '~')
                {
                if (buf[1] == 'a')
                    abort ();
                else
                    (void)raise (atoi (buf+1));
                }
            else if (buf[0] == '=')
                {
                n = atoi (buf+1);
                if (n >= 1 && n <= sizeof (buf) - 1)
                    req = n;
                else
                    req = sizeof (buf) - 1;
                }
            else if (buf[0] == '-')
                {
                n = fcntl (0, F_GETFL, 0);
                if (n >= 0)
                    {
                    n ^= O_NDELAY;
                    (void)fcntl (0, F_SETFL, n);
                    }
                }
            }
        }
    return (0);
    }
